package Apache::AuthRoles;

use Data::Dumper;
use Tie::DBI;
use strict;
use Apache::Constants ':common';

$Apache::AuthRoles::VERSION = '0.001';

sub handler {
  my $r = shift;

  # get user's authentication credentials
  my($res, $sent_pw) = $r->get_basic_auth_pw;
  return $res if $res != OK;
  my $user = $r->connection->user;

  my $reason = authenticate($r, $user, $sent_pw);

  if ($reason) {
    $r->note_basic_auth_failure;
    $r->log_reason($reason, $r->filename);
    return AUTH_REQUIRED;
  }
  return OK;
}


sub authenticate {
  my($r, $user, $sent_pw) = @_;

  $user && $sent_pw or return 'empty user names and passwords forbidden'; 

  # get configuration information
  my $dsn        = $r->dir_config('DB_dsn') || 'mysql:authen';
  my $dbuser     = $r->dir_config('DB_user') || '';
  my $dbpass     = $r->dir_config('DB_pass') || '';

  # tie db handles
  tie my %users, 'Tie::DBI', {
			      db    => $dsn,
			      table => 'users',
			      user  => $dbuser,
			      pass  => $dbpass,
			      key   => 'user',
			      CLOBBER => 0 # read only!
			     } or return "Could not connect to database $dsn!";

  tie my %groups, 'Tie::DBI', {
			       db    => $dsn,
			       table => 'groups',
			       user  => $dbuser,
			       pass  => $dbpass,
			       key   => 'groupname',
			       CLOBBER => 0 # read only!
			      } or return "Could not connect to database $dsn!";

  tie my %roles, 'Tie::DBI', {
			      db    => $dsn,
			      table => 'roles',
			      user  => $dbuser,
			      pass  => $dbpass,
			      key   => 'rolename',
			      CLOBBER => 0 # read only!
			     } or return "Could not connect to database $dsn!";

  tie my %group_members, 'Tie::DBI', {
				      db    => $dsn,
				      table => 'group_members',
				      user  => $dbuser,
				      pass  => $dbpass,
				      key   => 'idx',
				      CLOBBER => 0 # read only!
				     } or return "Could not connect to database $dsn!";

  tie my %role_members, 'Tie::DBI', {
				     db    => $dsn,
				     table => 'role_members',
				     user  => $dbuser,
				     pass  => $dbpass,
				     key   => 'idx',
				     CLOBBER => 0 # read only!
				    } or return "Could not connect to database $dsn!";

  my(@memberids,@req_role, @req_group);

  my $requires = $r->requires;
  for my $req (@$requires) {
    my($require, @list) = split /\s+/, $req->{requirement};
    if ($require eq "Role") {
      @req_role  = @list;
    }
    elsif ($require eq "Group") {
      @req_group = @list;
    }
  }


  # get all groups associated to the role(s), if any
  foreach my $role (@req_role) {
    next if(!$roles{$role});
    my @ids = (tied %role_members)->select_where("idx_role = $roles{$role}{idx}");
    foreach my $idx (@ids) {
      my @groups = (tied %groups)->select_where("idx = $role_members{$idx}{idx_group}");
      push @req_group, @groups if(@groups);
    }
  }

  if (!@req_group) {
    return "Invalid htaccess config, no(or not existent) role or group defined!";
  }


  $users{$user} or return "Access denied, invalid credentials!";


  # now see, if the user is a member in at least one of the possible groups!
  foreach my $group (@req_group) {
    next if(!$groups{$group});
    my @ids = (tied %group_members)->select_where("idx_user = $users{$user}{idx} AND idx_group = $groups{$group}{idx}");
    push @memberids, @ids if(@ids);
  }

  if (!@memberids) {
    return "Access denied by rule!";
  }

  # the actual auth :-)
  $users{$user}{passwd} eq crypt($sent_pw, $users{$user}{passwd}) or return "Access denied, invalid credentials!";

  # all well here? so return true
  return "";
}


1;

__END__

=head1 NAME

Apache::AuthRoles - mod_perl role/group/user based authentication module

=head1 SYNOPSIS

httpd.conf:

 <Location /secure>
   # load the module
   PerlAuthenHandler Apache::AuthRoles
   # Basic Realm
   AuthType          Basic
   # the database dsn (driver:database[:host=host:[port]])
   PerlSetVar        DB_dsn       mysql:authen
   PerlSetVar        DB_user      scip
   PerlSetVar        DB_pass      blah
 </Location>


per directory .htaccess:
 AuthName        "Challenging Experience"
 require         valid-user
 require         Role  Chefs Admins Staff
 require         Group Engineers

=head1 DESCRIPTION


This mod_perl based modules implements a role based authentication scheme for
apache basic authentication. It uses a DBI-capable database for roles/groups/users
  and it uses the B<Tie::DBI> DBI interface, which you can find on your next local
CPAN dealer.

You need to specify your .htaccess file one or more Roles and/or one or more Groups
which are allowed to access this directory. The user who tries to access your protected
directory must appear in at least one group which is a member under one of the defined
roles for this directory. If no Roles are specified but only group(s), then she must
be a member under one of these groups, right.


=head1 DATABASE

The database structure:

table B<users>:

 +--------+---------------+------+-----+---------+----------------+
 | Field  | Type          | Null | Key | Default | Extra          |
 +--------+---------------+------+-----+---------+----------------+
 | idx    | int(11)       |      | PRI | 0       | auto_increment |
 | ts     | timestamp(14) | YES  |     | NULL    |                |
 | user   | varchar(100)  |      | UNI |         |                |
 | passwd | varchar(100)  |      |     |         |                |
 +--------+---------------+------+-----+---------+----------------+

table B<groups>:

 +-----------+---------------+------+-----+---------+----------------+
 | Field     | Type          | Null | Key | Default | Extra          |
 +-----------+---------------+------+-----+---------+----------------+
 | idx       | int(11)       |      | PRI | 0       | auto_increment |
 | ts        | timestamp(14) | YES  |     | NULL    |                |
 | groupname | varchar(200)  |      | UNI |         |                |
 +-----------+---------------+------+-----+---------+----------------+

table B<roles>:

 +----------+---------------+------+-----+---------+----------------+
 | Field    | Type          | Null | Key | Default | Extra          |
 +----------+---------------+------+-----+---------+----------------+
 | idx      | int(11)       |      | PRI | 0       | auto_increment |
 | ts       | timestamp(14) | YES  |     | NULL    |                |
 | rolename | varchar(200)  |      | UNI |         |                |
 +----------+---------------+------+-----+---------+----------------+

table B<group_members>:

 +-----------+---------------+------+-----+---------+----------------+
 | Field     | Type          | Null | Key | Default | Extra          |
 +-----------+---------------+------+-----+---------+----------------+
 | idx       | int(16)       |      | PRI | 0       | auto_increment |
 | ts        | timestamp(14) | YES  |     | NULL    |                |
 | idx_user  | int(11)       |      | MUL | 0       |                |
 | idx_group | int(11)       |      |     | 0       |                |
 +-----------+---------------+------+-----+---------+----------------+

table B<role_members>:


 +-----------+---------------+------+-----+---------+----------------+
 | Field     | Type          | Null | Key | Default | Extra          |
 +-----------+---------------+------+-----+---------+----------------+
 | idx       | int(16)       |      | PRI | 0       | auto_increment |
 | ts        | timestamp(14) | YES  |     | NULL    |                |
 | idx_group | int(11)       |      | MUL | 0       |                |
 | idx_role  | int(11)       |      |     | 0       |                |
 +-----------+---------------+------+-----+---------+----------------+


=head1 AUTHOR

Thomas Linden <tom@daemon.de>


=head1 COPYRIGHT

Copyright (c) 2000 Thomas Linden

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.


=cut

